//
// Copyright(C) 2016-2018 Michael "Drahsid" Smith
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//

//A spring solver system, using float values. Generates the "smoothest" possible procedural motion
class Spring
{
	//Euler's number
	float e = 2.718281828459045f;

	//Damper, or the "bouncyness" or stability of the spring.
	float d=1;

	//Speed, or how quickly the point reaches the target, the stiffness.
	float s=1;

	//The current point of the spring.
	float p0 = 0;

	//The current target position of the spring.
	float p1 = 0;

	//The current velocity of the spring.
	float v0 = 0;

	//The target velocity of the spring.
	float v1 = 0;

	Spring(){}
	Spring(float Speed, float Damper)
	{
		s = Speed;
		d = Damper;
	}
	

	//Methods

	//Update the position and velocity of the spring, also returns the updated position.
	void posvel()
	{
		float x = GAME_DELTA_TIME;
		if (s==0)
		{
			p0=p0;
			v0=v0;
		}

		else if (d < 1 - 1e-8f)
		{
			float h = Math::Pow((1 - d * d), 0.5f);
			float c1 = (p0 - p1 + v1 * 2 * d / s);
			float c2 = (p0 - p1) * d / h + v0 / (h * s) + v1 * (2 * d * d - 1) / (h * s);
			float co = Math::Cos(h * s * x);
			float si = Math::Sin(h * s * x);
			float ex = Math::Pow(e, d * s * x);

			p0 = c1 * co / ex + c2 * si / ex + p1 + v1 * (x - 2 * d / s);
			v0 = c2 * s * (co * h - d * si) / ex - c1 * s * (co * d + h * si) / ex + v1;

		}
		
		else if (d > 1 - 1e-8f)
		{
			float c1 = p0 - p1 + v1 * 2 / s;
			float c2 = p0 - p1 + (v0 + v1) / s;
			float ex = Math::Pow(e, s * x);

			p0 = (c1 + c2 * s * x) / ex + p1 + v1 * (x - 2 / s);
			v0 = v1 - (c1 + (s * x - 1) * c2) * s / ex;
		}
		
		else
		{
			float h = Math::Pow((d * d - 1), 0.5f);
			float x0 = (v1 - v0) / (2 * s * h);
			float x1 = v1 * d / s - (p1 - p0) / 2;
			float c1 = x1 * (1 - d / h) + x0;
			float c2 = x1 * (1 + d / h) - x0;
			float co = Math::Pow(e, (-(h + d) * s * x));
			float si = Math::Pow(e, (h + d) * s * x);

			p0 = c1 * co + c2 * si + p1 + v1 * (x - 2 * d / s);
			v0 = c2 * si * (h - d) * s - c1 * co * (d + h) * s + v1;

		}
	}

	//Get an updated target position
	void Targpv()
	{
		p0 = p1 + (v1 * GAME_DELTA_TIME);
		v0 = v1;
	}

	//Public Stuff
	
	//The current point of the spring.
	float getPosition()
	{
		posvel();
		return p0;
	}
	float getPositionWithoutUpdate()
	{
		return p0;
	}
	void setPosition(float&in value)
	{
		p0 = value;
		posvel();
		Targpv();
	}

	//The current velocity of the spring.
	float getVelocity()
	{
		posvel();
		return v0;
	}
	void setVelocity(float&in value)
	{
		v0 = value;
		posvel();
	}

	//The current target position of the spring.
	float getTarget()
	{
		posvel();
		Targpv(); 
		return p1;
	}
	void setTarget(float&in value)
	{
		p1 = value;
		posvel();
	}

	//The target velocity of the spring.
	float getTargetVelocity() 
	{
		return v1;
	}
	void setTargetVelocity(float&in value)
	{
		v1 = value;
		posvel();
		Targpv();
	}

	//Damper, or the "bouncyness" or stability of the spring.
	float getDamper()
	{
		posvel();
		return d; 
	}
	void setDamper(float&in value)
	{
		d = value;
		posvel();
	}

	//Speed, or how quickly the point reaches the target, the stiffness.
	float getSpeed()
	{
		posvel();
		return s;
	}
	void setSpeed(float&in value)
	{
		s = value;
		posvel();
	}

	//Add acceleration to the spring.
	void Accelerate(float&in value)
	{
		posvel();
		Targpv();
		v0 += value;
	}
}

//Resolve weird math issues with different implementation
class VecSpring
{
	Spring x;
	Spring y;
	Spring z;

	float d=1;
	float s=1;

	VecSpring()
	{
		x = Spring(s, d);
		y = Spring(s, d);
		z = Spring(s, d);
	}
	VecSpring(float Speed, float Damper)
	{
		s = Speed;
		d = Damper;

		x = Spring(s, d);
		y = Spring(s, d);
		z = Spring(s, d);
	}

	kVec3 getPosition()
	{
		return kVec3(x.getPosition(), y.getPosition(), z.getPosition());
	}
	kVec3 getPositionWithoutUpdate()
	{
		return kVec3(x.getPositionWithoutUpdate(), y.getPositionWithoutUpdate(), z.getPositionWithoutUpdate());
	}
	void setPosition(kVec3&in p)
	{
		x.setPosition(p.x);
		y.setPosition(p.y);
		z.setPosition(p.z);
	}

	kVec3 getTarget()
	{
		return kVec3(x.getTarget(), y.getTarget(), z.getTarget());
	}
	void setTarget(kVec3&in t)
	{
		x.setTarget(t.x);
		y.setTarget(t.y);
		z.setTarget(t.z);
	}

	kVec3 getVelocity()
	{
		return kVec3(x.getVelocity(), y.getVelocity(), z.getVelocity());
	}
	void setVelocity(kVec3&in v)
	{
		x.setVelocity(v.x);
		y.setVelocity(v.y);
		z.setVelocity(v.z);
	}

	kVec3 getTargetVelocity()
	{
		return kVec3(x.getTargetVelocity(), y.getTargetVelocity(), z.getTargetVelocity());
	}
	void setTargetVelocity(kVec3&in v)
	{
		x.setTargetVelocity(v.x);
		y.setTargetVelocity(v.y);
		z.setTargetVelocity(v.z);
	}

	float getSpeed() { return s; }
	void setSpeed(float&in Speed)
	{
		s = Speed;
		x.setSpeed(s);
		y.setSpeed(s);
		z.setSpeed(s);
	}

	float getDamper() { return d; }
	void setDamper(float&in Damper)
	{
		d = Damper;
		x.setDamper(d);	
		y.setDamper(d);	
		z.setDamper(d);	
	}

}
